# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

""""Abstracting out authentication - this is basically a minimal implementation."""

import logging
from base64 import b64encode


class Authenticate:

    def __init__(self, identity, keys):
        # Create an authentication object when passed location and (our) keys
        self.identity = identity
        self.keys = keys

    def auth(self, msg):
        # Authenticate against config
        # Returns success, user, config
        try:
            pk = b64encode(msg.params['user']).decode()  # the database holds base64 encoded strings for sanity
        except TypeError:
            return False, False, None

        # A user?
        try:
            email, json = self.identity.user_config_from_db(pk)
            logging.info("Authentication succeeded as user: " + email)
            return True, True, json
        except ValueError:  # throws if not a user (perhaps a node)
            pass

        # A node, then?
        try:
            json = self.identity.node_config_from_db(pk)
            logging.info("Authentication succeeded as node: " + str(pk))
            return True, False, json
        except ValueError:  # oh, we failed then
            pass

        # bugger
        logging.warning("Authentication failed - user not in database (and is not a node): " + str(pk))
        return False, False, None

    def __repr__(self):
        return "<messidge.broker.auth.Auth object at %x (pk=%s)>" % (id(self), self.keys.public)



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import libnacl
import libnacl.utils
import cbor
import psutil
import time
from binascii import hexlify
from threading import Thread
from _thread import allocate_lock, get_ident
from base64 import b64decode
import shortuuid
import zmq
from zmq.utils.monitor import recv_monitor_message
from .message import Message
from messidge import KeyPair, Waitable
from messidge.loop import Loop


class Connection(Waitable):
    """Connection onto Messidge broker."""

    # Is expecting ~/.messidge/ to contain keys named after the location to connect to (eg)
    # -rw-r--r--  1 dpreece  staff   45B Jul  8 16:36 localhost
    # -r--------  1 dpreece  staff   45B Jul  8 16:36 localhost.pub

    # skt is the main tcp socket from here to the location and belongs to the background thread
    # x_thread_socket is the "pickup" end of the per-thread sending sockets forwards to skt
    # Note that messages received are exclusively dealt with on the background thread

    def __init__(self, location: str=None, *, prefix='~/.messidge',
                 exit_on_exception=False, reflect_value_errors=False,
                 location_ip: str=None, keys: KeyPair=None, server_pk: str=None):
        """Instantiate a connection.

        :param location: The FQDN of the location to connect to.
        :param prefix: Directory for the client keys and server public keys.
        :param exit_on_exception: Causes the message loop to exit if there's an uncaught exception.
        :param reflect_value_errors: If True, raised ValueErrors will have their source messages replied to.
        :param location_ip: An override for the DNS resolution of 'location'. Useful for LAN and/or test connections.
        :param keys: An override for the key pair in the 'prefix' directory.
        :param server_pk: An override for the server public key in the 'prefix' directory."""
        super().__init__()
        self.connect_ip = location_ip if location_ip is not None else location
        self.location = location if location is not None else self.connect_ip
        self.keys = keys if keys is not None else KeyPair(location, prefix=prefix)
        self.inproc_name = "inproc://x_thread/" + str(id(self))
        self.session_key = None
        self.connect_callbacks = set()
        self.reflect_value_errors = reflect_value_errors
        self.exit_on_exception = exit_on_exception
        self.thread_skt = {}  # map from thread to the socket used to send messages on it's behalf
        self.rid = b''
        self.loop = None
        self.loop_block = allocate_lock()
        self.uuid_blockreply = {}
        self.uuid_blockresults = {}
        self.connected = False
        self.loop_thread = None

        # fetch the server's public key
        if server_pk is None:
            try:
                filename = os.path.expanduser(prefix + '/%s.spub' % location)
                with open(filename, 'r') as f:
                    server_pk = f.read().strip('\n')
            except FileNotFoundError:
                raise RuntimeError("There is no server public key in %s so cannot connect" % filename)
        self.server_public_binary = b64decode(server_pk)
        if len(self.server_public_binary) != 32:
            raise RuntimeError("The server public key record is broken.")

        # Threading
        zmq.Context.instance().set(zmq.IO_THREADS, psutil.cpu_count())
        logging.debug("ZeroMQ context IO threads: " + str(zmq.Context.instance().get(zmq.IO_THREADS)))

        # Prevent queued messages being sent instead of a handshake when reconnecting
        zmq.Context.instance().setsockopt(zmq.LINGER, 0)

        # Should be able to connect, then.
        self.loop_block.acquire()  # allows the thread to start and prepare but not actually run, yet.
        self.thread = Thread(target=self._start, name="Messidge background loop")
        self.thread.start()

    def start(self):
        """Start message loop - separate from __init__ so we get a chance to register_exclusive/register_commands"""
        self.loop_block.release()  # causes the block in _start to clear
        return self

    def wait_until_complete(self):
        """Blocks until the message loop exits"""
        self.thread.join()
        if self.exception is not None:  # re-raises (on the main thread) if the loop caught an exception
            raise self.exception

    def disconnect(self):
        """Stop the message loop and disconnect - without this object cannot be garbage collected"""
        self.loop.stop()
        self.thread.join()
        for thread_id in list(self.thread_skt.keys()):
            self.destroy_send_skt_for(thread_id)
        self.thread_skt.clear()
        self.connect_callbacks.clear()
        self.uuid_blockreply.clear()
        self.uuid_blockresults.clear()

    def _start(self):
        """The message loop runs on a background thread"""
        self.loop_thread = get_ident()

        # create the trunk socket - remember sockets must be created on the thread they are used on
        logging.info("Connecting to: %s:2020" % self.location)
        working = False
        while not working:
            try:
                self.skt = zmq.Context.instance().socket(zmq.DEALER)
                self.skt.setsockopt(zmq.LINGER, 0)
                self.skt.connect("tcp://%s:2020" % self.connect_ip)
                self.skt_monitor = self.skt.get_monitor_socket()
                working = True
            except zmq.error.ZMQError:
                del self.skt
                self.skt = None
                logging.info("Retrying ZMQ socket")
                time.sleep(1)
        logging.debug("Trunk socket is: %x" % id(self.skt))

        # create the cross thread socket for forwarding
        # since there can be more than one connection in the same process we need to give it a unique id
        self.x_thread_receive = zmq.Context.instance().socket(zmq.DEALER)
        self.x_thread_receive.bind(self.inproc_name)
        logging.debug("Cross-thread socket is: %x" % id(self.x_thread_receive))

        # kick off a message loop - has to be constructed on background thread
        self.loop = Loop(self.skt, exit_on_exception=self.exit_on_exception)
        self.loop.register_exclusive(self.skt_monitor, self._socket_event, comment="Socket events")
        self.loop.register_exclusive(self.x_thread_receive, self._fast_forward, comment="Cross thread socket")
        if self.reflect_value_errors:
            self.loop.on_value_error(self._background_thread_exception)

        # block until allowed to go by start()
        self.loop_block.acquire()
        self.mark_as_ready()
        self.loop.run()  # blocks until loop.stop is called or the loop catches a wayward exception
        self.exception = self.loop.caught_exception

        # Maybe the main thread was in wait_until_ready
        if self.exception is not None:
            self.unblock_and_raise(self.exception)

        # Maybe the main thread was in send_blocking_cmd
        for block in self.uuid_blockreply.values():
            block.release()

    def _socket_event(self, monitor_socket):
        # is this a connection event?
        event = recv_monitor_message(monitor_socket)
        if event['event'] == zmq.EVENT_DISCONNECTED:
            if len(self.uuid_blockreply) != 0:
                raise RuntimeError("Was disconnected from location with a blocking call still pending")
            logging.info("Have been disconnected from location. Wait...")
            self.connected = False
            return
        if event['event'] != zmq.EVENT_CONNECTED:
            return

        logging.info("Message queue connected")
        self.connected = True

        # send the encryption session request
        # the rid is used to show which session we *were* if reconnecting
        if self.rid != b'':
            logging.debug("Connecting with old rid: " + hexlify(self.rid).decode())
        params = {'user': self.keys.public_binary(), 'rid': self.rid}
        parts = [b'', b'auth', b'', params, b'']
        binary = cbor.dumps(parts)
        self.skt.send(binary)
        try:
            # poll so we can effect a timeout
            attempts = 10
            while True:
                try:
                    in_binary = self.skt.recv(flags=zmq.NOBLOCK)
                    attempts -= 1
                    if attempts == 0:
                        raise ValueError("Could not connect.")
                    break
                except zmq.ZMQError:
                    time.sleep(1)

            # did we actually authenticate?
            in_parts = cbor.loads(in_binary)
            if len(in_parts) != 3:
                raise ValueError("Authentication failed.")
            recv_nonce, enc_session_key, rid = in_parts
        except ValueError as e:
            self.unblock_and_raise(e)  # raises within a blocked wait_until_ready
            return

        # unwrap and set the session key
        try:
            if enc_session_key != b'':
                self.session_key = libnacl.crypto_box_open(enc_session_key, recv_nonce,
                                                           self.server_public_binary, self.keys.secret_binary())
                self.loop.set_crypto_params(self.session_key)
        except libnacl.CryptError:
            raise ValueError("Handshake failed.")
        logging.debug("Setting rid to: " + hexlify(rid).decode())
        self.rid = rid
        logging.info("Handshake completed")
        for callback in self.connect_callbacks:
            callback(self.rid)
        self.mark_as_ready()

    def send_skt(self):
        """Allocates (if necessary) a socket for the calling thread to send messages with."""
        thread = get_ident()

        # if this is the loop thread then we can use the main thread socket to send
        if thread is self.loop_thread:
            return self.skt

        try:
            return self.thread_skt[thread]
        except KeyError:
            # need a new socket
            new_skt = zmq.Context.instance().socket(zmq.DEALER)
            new_skt.connect(self.inproc_name)
            self.thread_skt[thread] = new_skt
            # logging.debug("Created a per-thread socket for: " + str(thread))
            return new_skt

    def destroy_send_skt(self):
        """Closes and removes the send_skt for the calling thread."""
        self.destroy_send_skt_for(get_ident())

    def destroy_send_skt_for(self, thread):
        try:
            self.thread_skt[thread].close()
            del self.thread_skt[thread]
            # logging.debug("Destroyed per-thread socket for: " + str(thread))
        except KeyError:
            pass

    def send_cmd(self, cmd: bytes, params=None, bulk: bytes=b'', uuid=b'', reply_callback=None):
        """Sends a command to the location, can route replies.

        :param cmd: the command.
        :param params: A {'key': 'value'} dictionary of parameters or ['list'].
        :param bulk: An optional piece of bulk data to transport.
        :param uuid: A uuid to attach to this command (so it can reply).
        :param reply_callback: Callback to fire when the command receives a reply - gets passed the message.
        """
        # BS check
        if self.loop is None:
            raise RuntimeError("The connection has no message loop - _start has not been called.")
        if cmd is None:
            raise RuntimeError("Need to pass a command to send_cmd.")

        # send
        if reply_callback is not None:
            # important that we register the expectation of a reply before asking the question
            if uuid is b'':
                uuid = shortuuid.uuid().encode()  # as bytes
            self.loop.register_reply(uuid, reply_callback)

        Message.send(self.send_skt(), cmd, self.session_key, params, uuid=uuid, bulk=bulk)

    def send_blocking_cmd(self, cmd: bytes, params=None, bulk: bytes=b'', timeout: float=240) -> Message:
        """Sends a command to the location and blocks waiting for a reply (which is returned).
        May raise ValueError exceptions.

        :param cmd: the command.
        :param params: A {'key': 'value'} dictionary of parameters or ['list'].
        :param bulk: An optional piece of bulk data to transport.
        :param timeout: In seconds.
        :return: The reply message.
        """
        thread = get_ident()
        uuid = shortuuid.uuid().encode()

        # if on the background thread we need to block differently...
        if thread != self.loop_thread:
            # acquire a lock, wait for the reply
            self.uuid_blockreply[uuid] = allocate_lock()
            self.uuid_blockreply[uuid].acquire()
            self.send_cmd(cmd, params, uuid=uuid, reply_callback=self._unblock, bulk=bulk)

            # when the background thread has an answer, the lock will release and we can continue
            if not self.uuid_blockreply[uuid].acquire(timeout=timeout):
                raise ValueError("Blocking call timed out: %s(%s)" % (cmd.decode(), str(params)))
            msg = self.uuid_blockresults[uuid]

            # clean up
            self.uuid_blockreply[uuid].release()
            del self.uuid_blockreply[uuid]
            del self.uuid_blockresults[uuid]

        # background thread
        else:
            self.loop.register_reply(uuid, None)
            self.send_cmd(cmd, params, uuid=uuid, bulk=bulk)
            msg = self.loop.run()
            self.loop.unregister_reply(uuid)

        # raise exceptions if either in the message or caught on a background thread
        if 'exception' in msg.params:
            raise ValueError(msg.params['exception'])
        return msg

    def register_commands(self, obj, commands):
        """Register a list of commands to be handled by the loop.

        :param obj: The object that will handle the commands.
        :param commands: A dict of the form {b'command': handler, ...}."""
        while self.loop is None:
            os.sched_yield()  # loop creation has not yet had a timeslice
        self.loop.register_commands(self.skt, obj, commands)

    def register_connect_callback(self, callback):
        """Register a callback to be fired once the connection is complete.

        :param callback: the object.method to call - is passed the connection rid."""
        self.connect_callbacks.add(callback)

    def unregister_connect_callback(self, callback):
        """Unregister a callback to be fired once the connection is complete.

        :param callback: the object.method to cancel."""
        try:  # may have already been disconnected as part of object shutting down
            self.connect_callbacks.remove(callback)
        except KeyError:
            pass

    def location_name(self) -> str:
        """Outside access to the name/address of this location.

        :return: The FQDN of the location as a string."""
        return self.location

    def _fast_forward(self, skt):
        binary = skt.recv()
        self.skt.send(binary)

    def _unblock(self, msg: Message):
        try:
            self.uuid_blockresults[msg.uuid] = msg  # msg is stored first
            self.uuid_blockreply[msg.uuid].release()
            self.loop.unregister_reply(msg.uuid)
        except KeyError:
            logging.error("Message from a blocking call arrived late (ignored): " + str(msg))

    def _background_thread_exception(self, e, msg):
        """Call from non-main threads to return a value error to the client"""
        if msg.replyable():
            msg.reply(self.send_skt(), {"exception": str(e)})
            logging.info("Client called %s and raised a ValueError: %s" % (msg.command.decode(), str(e)))
        else:
            logging.warning("Unable to return ValueError to client: " + str(e))

    def __repr__(self):
        return "<messidge.client.connection.Connection object at %s (location=%s)>" % (id(self), self.location)


def cmd(required_params, *, needs_reply=False):
    """Create the internal structure describing a command

    :param required_params: A list of parameters that must be included with the command.
    :param needs_reply: The message needs to be replied to (and must have a uuid)."""
    return required_params, needs_reply



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# This is the top level object and message broker

# Note that commands that get called and raise a ValueError have the exception passed back to the caller


import logging
import random
import libnacl.utils
from base64 import b64encode
from binascii import hexlify
import os

import zmq
import cbor
from zmq.utils.monitor import recv_monitor_message
from .identity import Identity, AccountConfirmationServer
from .agent import Agent
from .auth import Authenticate

from .. import KeyPair
from ..loop import Loop
from .message import BrokerMessage


class Broker:
    """The broker is the top level object for creating a server/broker."""

    def __init__(self, keys: KeyPair, model, node_type, session_type, controller, *, base_port: int=2020,
                 identity_type=Identity, auth_type=Authenticate, account_confirm_type=AccountConfirmationServer,
                 pre_run_callback=None, node_create_callback=None, node_destroy_callback=None,
                 session_recovered_callback=None, session_destroy_callback=None,
                 forwarding_insert_callback=None, forwarding_evict_callback=None):
        """:param keys: Server public/secret keys.
        :param model: Model object (derived from ModelMinimal).
        :param node_type: Class to construct node objects with (derived from NodeMinimal).
        :param session_type: Class to construct session objects with (derived from SessionMinimal).
        :param controller: Controller object (derived from ControllerMinimal).
        :param base_port: TCP port to use plus the next one (so default is 2020 and 2021).
        :param identity_type: Potentially custom class to construct an identity server (see identity.py).
        :param auth_type: Class to authenticate with.
        :param account_confirm_type: Class to use for confirming client accounts.
        :param pre_run_callback: Fired last thing before the message loop starts.
                                 ...signature ()
        :param node_create_callback: Fired once a new node has been created.
                                     ..signature (pk)
        :param node_destroy_callback: Fired last thing before a node is destroyed (was disconnected, say).
                                      ...signature (pk)
        :param session_recovered_callback: Fired with the new rid when a session reconnects.
                                           ...signature (session, old_rid, new_rid)
        :param session_destroy_callback: Fired last thing before a session is destroyed.
                                         ...signature (rid)
        :param forwarding_insert_callback: Called when a new entry is made into the forwarding map.
                                           ...signature (key, value)
        :param forwarding_evict_callback: Called when an entry is removed from the forwarding map.
                                          ...signature (key, value)

        The eviction of an entry from the map is not to be used as an indication that an object has been destroyed,
        it may be considerably after destruction or not at all.
        """
        # Basic structure
        self.keys = keys
        self.model = model
        self.controller = controller
        self.commands = controller.commands
        self.node_type = node_type
        self.session_type = session_type
        self.session_recovered_callback = session_recovered_callback
        self.session_destroy_callback = session_destroy_callback
        self.node_create_callback = node_create_callback
        self.node_destroy_callback = node_destroy_callback
        self.pre_run_callback = pre_run_callback
        self.base_port = base_port
        self.skt = None  # initialised when run
        self.skt_monitor = None
        self.fd_rid = {}  # map from file descriptor to remote id so we know who has disconnected
        self.rid_pk = {}
        self.rid_session_key = {}
        self.node_rid_pk = {}
        self.node_pk_rid = {}
        self.short_term_forwards = {}
        self.forwarding_insert_callback = forwarding_insert_callback
        if forwarding_evict_callback:
            model.long_term_forwards.set_callback(forwarding_evict_callback)
        self.identity = identity_type()
        self.authenticator = auth_type(self.identity, self.keys)
        self.account_confirm = account_confirm_type(self.identity, self.keys, base_port + 1)
        self.agent = Agent()
        self.loop = None
        self.next_connection_rid = None

        # Threading
        zmq.Context.instance().set(zmq.IO_THREADS, os.cpu_count())
        logging.debug("ZeroMQ context IO threads: " + str(zmq.Context.instance().get(zmq.IO_THREADS)))

    def run(self):
        """Call on the main thread, creates the background thread that actually does all the work."""
        # the server socket
        self.skt = zmq.Context.instance().socket(zmq.ROUTER)
        self.skt.set_hwm(0)  # infinite, don't drop or block because that would be bad
        self.skt.setsockopt(zmq.LINGER, 0)  # drop waiting packets if a client disconnects
        try:
            self.skt.bind("tcp://*:" + str(self.base_port))
        except zmq.error.ZMQError:
            logging.error("Cannot bind socket - is there an instance of the broker running already?")
            raise RuntimeError("Cannot bind zmq listen socket, exiting")
        self.set_next_rid()

        # Set up the message loop
        self.loop = Loop(self.skt, message_type=BrokerMessage, exit_on_exception=True)
        self.loop.register_exclusive(self.skt, self._event_socket, comment="ZMQ")
        self.loop.register_exclusive(self.agent.decrypt_pipe[0].fileno(), self._event_pipe, comment="Decrypted")
        self.loop.register_exclusive(self.agent.encrypt_pipe[0].fileno(), self._emit_encrypted, comment="Encrypted")
        self.loop.register_exclusive(self.skt.get_monitor_socket(), self._monitor, comment="Events Monitor")

        # Client hookups...
        if self.pre_run_callback is not None:
            self.pre_run_callback()

        # Go
        logging.info("Started broker: 0.0.0.0:" + str(self.base_port))
        self.loop.run()

    def stop(self):
        """Stop background threads. Must be called to allow garbage collection and for a clean exit."""
        # DO NOT explicitly disconnect user sessions or nodes

        # stop the loop
        if self.loop is not None:  # if can't bind socket, bins out before loop is constructed
            self.loop.stop()

        # stop any threads
        [obj.stop() for obj in (self.identity, self.agent, self.account_confirm)]

    def send_cmd(self, rid, command: bytes, params: dict, *, bulk: bytes=b'', uuid: bytes= b''):
        """Send command to either a node or user

        :param rid: the routing identifier to send the message to.
        :param command: a byte string of the command to send.
        :param params: A {'key': 'value'} dictionary of parameters to send.
        :param bulk: An optional piece of bulk data to transport.
        :param uuid: A uuid to attach to this command (so it can reply).
        """
        try:
            session_key = self.rid_session_key[rid]
        except KeyError:
            logging.debug("Failed sending command, no session_key for rid: " + hexlify(rid).decode())
            # don't call 'disconnect' on the rid because
            # this might be a recovering session that hasn't negotiated yet
            return

        # session_key is None implies unencrypted
        if session_key is None:
            BrokerMessage.send_socket(self.skt, rid, b'', command, uuid, params, bulk)
        else:
            BrokerMessage.send_pipe(self.agent.encrypt_pipe[0], rid, session_key, b'',
                                    command, uuid, params, bulk)

    def set_next_rid(self):
        """Set the routing id for the next socket that connects."""
        self.next_connection_rid = bytes([random.randint(0, 255) for _ in range(0, 4)])
        logging.debug("Set next rid to: " + hexlify(self.next_connection_rid).decode())
        self.skt.set(zmq.CONNECT_RID, self.next_connection_rid)

    def disconnect_for_rid(self, rid):
        """Disconnect a client.

        :param rid: routing id for the client to disconnect."""
        # Remove the crypto agent
        session_key = None
        try:
            # both users and nodes have rid_session_key mappings (nodes map to None)
            session_key = self.rid_session_key[rid]
            del self.rid_session_key[rid]
        except KeyError:
            logging.debug("While disconnecting rid could not find session key or None: " + hexlify(rid).decode())
            try:
                del self.model.sessions[rid]
                self.model.delete_session_record(rid)
            except KeyError:
                pass
            return

        # Remove the right objects
        if session_key is not None:
            self._disconnect_session(rid)
        else:
            self._disconnect_node(rid)

    def _handshake(self, msg, skt):
        # gets called via the event loop just like everything else
        # because this is the handshake message it's not actually encrypted
        # NOTE that if this falls over the client may/will become jammed and unable to reconnect
        if msg.command != b'auth':
            logging.debug("Not a handshake message on session: " + hexlify(msg.rid).decode())
            return False, None

        # authenticate
        success, user, config = self.authenticator.auth(msg)
        if not success:
            logging.warning("There was a protocol-correct failure to authenticate: " +
                            b64encode(msg.params['user']).decode())
            return False, None

        # create the session - the 'create session' calls reply to the handshake message
        if user:
            return True, self._create_user_session(msg, skt, config)
        else:
            rtn = True, self._create_node_session(msg, skt, config)
            if self.node_create_callback is not None:
                self.node_create_callback(msg.params['user'])
            return rtn

    def _create_user_session(self, msg, skt, config):
        # are we reconnecting to a persisted session? params['rid'] refers to the rid it *used* to be
        # tag words: recover, rewrite, map
        pk = msg.params['user']
        resources = None
        if msg.params['rid'] in self.model.sessions:  # pre-existing session
            # Update rids so the session now has the new rid
            session = self.model.sessions[msg.params['rid']]
            logging.info("Recovering session: %s -> %s" %
                         (hexlify(msg.params['rid']).decode(), hexlify(msg.rid).decode()))
            del self.model.sessions[msg.params['rid']]
            self.model.delete_session_record(msg.params['rid'])
            session.rid = msg.rid  # so now the session "belongs" to the new rid
            for fd, old_rid in self.fd_rid.items():
                if old_rid == msg.params['rid']:
                    self.fd_rid[fd] = msg.rid
            self.model.sessions[msg.rid] = session
            self.model.create_session_record(session)  # don't factor out, the callback needs to be called AFTER
            if self.session_recovered_callback is not None:
                self.session_recovered_callback(session, msg.params['rid'], msg.rid)
        else:
            # new session
            session = self.session_type(msg.rid, pk)
            self.model.sessions[msg.rid] = session
            self.model.create_session_record(session)
            resources = self.model.resources(pk)  # only create for new session

        session_key = libnacl.utils.salsa_key()
        logging.info("Connected session: " + hexlify(msg.rid).decode())
        self.rid_session_key[msg.rid] = session_key
        self.rid_pk[msg.rid] = pk

        # send the session key to the other end
        nonce = libnacl.utils.rand_nonce()
        enc_session_key = libnacl.crypto_box(session_key, nonce, pk, self.keys.secret_binary())
        parts = [nonce, enc_session_key, msg.rid]
        binary = cbor.dumps(parts)
        skt.send_multipart((msg.rid, binary))

        # maybe create a resource offer
        if resources is not None:
            BrokerMessage.send_pipe(self.agent.encrypt_pipe[0], msg.rid, session_key,
                                    b'', b'resource_offer', b'', resources)

        return session_key

    def _create_node_session(self, msg, skt, config):
        pk = msg.params['user']
        # this might be a node reconnecting
        try:
            old_rid = self.node_pk_rid[pk]  # will throw a key error if this is a fresh connection
            del self.node_rid_pk[old_rid]  # pk_rid gets overwritten so doesn't need deleting
            self.node_rid_pk[msg.rid] = pk
            self.node_pk_rid[pk] = msg.rid
            del self.rid_session_key[old_rid]
            logging.info("Reconnected node: " + b64encode(pk).decode())
        except KeyError:
            # a node connecting for the first time
            node = self.node_type(pk, msg, config)
            self.model.nodes[pk] = node
            self.node_rid_pk[msg.rid] = pk
            self.node_pk_rid[pk] = msg.rid
            logging.info("Connected node pk: " + b64encode(pk).decode())

        # send a blank session key so the connection knows it's not encrypted
        self.rid_session_key[msg.rid] = None
        parts = [b'', b'', msg.rid]
        binary = cbor.dumps(parts)
        skt.send_multipart((msg.rid, binary))
        return None

    def _monitor(self, skt):
        # Catches connect and disconnect messages from zmq
        evt = recv_monitor_message(skt)
        descriptor = evt['value']
        if evt['event'] == zmq.EVENT_ACCEPTED:  # the handshake will sort itself out
            logging.debug("Received a connection: %s -> %s" % (descriptor, hexlify(self.next_connection_rid).decode()))
            self.fd_rid[descriptor] = self.next_connection_rid

        # disconnection
        if evt['event'] == zmq.EVENT_DISCONNECTED:
            if descriptor not in self.fd_rid:
                logging.warning("Caught disconnection from a fd that has no connection to a rid")
                return
            else:
                logging.debug("Caught disconnection: " + str(descriptor))

            rid = self.fd_rid[descriptor]
            del self.fd_rid[descriptor]
            self.disconnect_for_rid(rid)

    def _disconnect_node(self, rid):
        # removing a node
        pk = b''
        try:
            pk = self.node_rid_pk[rid]
            del self.node_rid_pk[rid]
            del self.node_pk_rid[pk]
            del self.model.nodes[pk]
            if self.node_destroy_callback is not None:
                self.node_destroy_callback(pk)
        except KeyError:
            logging.debug("Closing node not held in the model: " + hexlify(rid).decode())
            return

        logging.info("Node disconnected: " + b64encode(pk).decode())

    def _disconnect_session(self, rid):
        # Disconnect a single session - fires a callback
        try:
            sess = self.model.sessions[rid]
            sess.close(self)
            del self.model.sessions[rid]
            del self.rid_pk[rid]
        except KeyError:
            logging.debug("Closing session not held in the model: " + hexlify(rid).decode())
            return

        if self.session_destroy_callback is not None:
            self.session_destroy_callback(rid)
        logging.info("Session disconnected: " + hexlify(rid).decode())

    # Entrypoint for events that have come in over zmq
    def _event_socket(self, skt):
        # receive from the socket
        try:
            msg = BrokerMessage.receive_socket(skt)
        except BaseException as e:
            logging.info("Something bad happened with a message coming over a socket: " + str(e))
            return

        # it will be plaintext if it came from a node
        try:
            session_key = self.rid_session_key[msg.rid]  # may be None, but still a valid (node) connection
            if session_key is None:  # a node
                msg.emit_socket = skt
            else:  # a user
                # send to be decrypted
                msg.session_key = session_key
                msg.forward_pipe(self.agent.decrypt_pipe[0])
                return

        # If a new connection.....
        except KeyError:
            self.set_next_rid()
            success, session_key = self._handshake(msg, skt)
            if not success:  # authentication failed
                msg.params = dict()
                msg.reply()
                logging.info("Disconnecting rid: " + hexlify(msg.rid).decode())
                self.disconnect_for_rid(msg.rid)
                return
            msg.session_key = session_key
            self.rid_session_key[msg.rid] = session_key  # may be None but marks the rid as being valid anyway
            return

        self._event(msg)

    def _event_pipe(self, fd):  # called when pipe 'fd' has a message it wants to inject into the event loop (decrypted)
        # receive from the pipe
        try:
            msg = BrokerMessage.receive_pipe(self.agent.decrypt_pipe[0], encrypted=False,
                                             reply_through=self.agent.encrypt_pipe[0])
        except BaseException as e:
            logging.info("Something bad happened with a message coming over a pipe: " + str(e))
            return

        # this may be a cry for help from the encryption agent
        if msg.params is None and msg.bulk is None:
            logging.error("Encryption agent could not decrypt")
            self.disconnect_for_rid(msg.rid)
            return

        # all good
        self._event(msg)

    def _event(self, msg):
        # at this point we know the message is plaintext and will have either emit_pipe or emit_socket set
        # ensure we have all the necessary parameters then forward, call the handler, validate, whatever
        try:
            # we can only have params be a dict at this point
            if not isinstance(msg.params, dict):
                raise ValueError("Message parameters must be a dictionary.")

            # is this a reply we're supposed to be forwarding?
            if msg.uuid in self.short_term_forwards or msg.uuid in self.model.long_term_forwards:
                try:
                    msg.rid = self.short_term_forwards[msg.uuid]
                    del self.short_term_forwards[msg.uuid]
                    # don't return here because it may yet promote into a long term reply
                except KeyError:
                    pass  # must be in the long term reply map

                # does this message want to be part of a long term reply?
                if msg.command == b'ka':  # i.e. keep-alive
                    msg.command = b''
                    if msg.rid in self.rid_session_key:
                        if msg.uuid not in self.model.long_term_forwards:
                            logging.debug("Start forwarding: %s -> %s" % (msg.uuid.decode(), hexlify(msg.rid).decode()))
                            self.model.long_term_forwards[msg.uuid] = msg.rid
                            self.forwarding_insert_callback(msg.uuid, msg.rid)  # persists
                        msg.rid = self.model.long_term_forwards[msg.uuid]
                        logging.debug("Forwarded to: " + hexlify(msg.rid).decode())
                    else:
                        logging.debug("Message wanted forwarding but no session: " + hexlify(msg.rid).decode())
                        return

                # if forwarding to a session, try to do so now
                if msg.rid not in self.node_rid_pk:
                    try:
                        msg.session_key = self.rid_session_key[msg.rid]
                        msg.forward_pipe(self.agent.encrypt_pipe[0])
                    except KeyError:
                        logging.debug("Wanted to send a message reply to a disconnected session: " + msg.uuid.decode())
                else:
                    # message will be forwarded to a node in plaintext
                    msg.forward_socket(self.skt)

                return

            # now we're either dealing with a command or forwarding it to the node so let's add some context
            if msg.command.decode() != "update_stats":
                logging.debug('Event: ' + msg.command.decode())
            if msg.rid in self.node_rid_pk:
                msg.params['node'] = self.node_rid_pk[msg.rid]
                # logging.debug("...apparently came from node: " + b64encode(msg.params['node']).decode())
            elif msg.rid in self.rid_pk and self.rid_pk[msg.rid] is not None:
                msg.params['user'] = self.rid_pk[msg.rid]
                msg.params['session'] = msg.rid
                # logging.debug("...apparently came from session: " + hexlify(msg.rid).decode())

            # if this is a message that laksa can deal with, call the controller
            if msg.command in self.commands:
                self._check_basic_properties(msg)
                getattr(self.controller, '_' + msg.command.decode())(msg)
                return

            # forwarding to the node
            try:
                # cache node_pk (don't need to send this to the node itself)
                node_pk = msg.params['node']
                del msg.params['node']

                # record the source rid so we can forward replies
                if msg.replyable():
                    self.short_term_forwards[msg.uuid] = msg.rid

                # set the rid to the node and forward
                msg.rid = self.node_pk_rid[node_pk]
                msg.forward_socket(self.skt)

            except KeyError:
                raise ValueError("No node address or invalid: " + str(msg.command))

        except ValueError as e:
            logging.info("Client %s called %s and raised a ValueError: %s" %
                         (hexlify(msg.rid).decode(), msg.command.decode(), str(e)))
            try:
                msg.reply({"exception": str(e)})
            except BaseException as e:
                logging.error("Failed while attempting to forward a value error - str(e) failed.")
                raise e

    def _check_basic_properties(self, msg):
        # Helper utility to bounce missing/clearly-wrong properties before they do a bad thing

        # is it expecting a reply and can we send it?
        try:
            try:
                necessary_params, needs_reply, node_only = self.commands[msg.command]
            except ValueError as e:
                raise RuntimeError("Problem unpacking the command structure - you have all three values?: " + str(e))
            if needs_reply and not msg.replyable():
                raise ValueError("Message needs to be replyable for: " + str(msg.command))
            if node_only and msg.rid not in self.node_rid_pk:
                raise ValueError("This call is for nodes only")
        except KeyError:
            # not a broker command, assume passing through (node value has already been checked)
            return  # no need to check params, we know message is not for the broker

        # have all the parameters for the command been sent?
        for necessary in necessary_params:
            if necessary not in msg.params:
                raise ValueError("Necessary parameter was not passed: " + necessary)

    def _emit_encrypted(self, fd):
        # called when pipe 'fd' has an encrypted message ready to send
        try:
            rid, session_key, parts = self.agent.encrypt_pipe[0].recv()
        except EOFError:
            logging.error("EOF while receiving in _emit_encrypted: " + str(fd))
            return
        binary = cbor.dumps(parts)
        self.skt.send_multipart((rid, binary))

    def __repr__(self):
        return "<messidge.broker.Broker object at %x>" % id(self)


def cmd(required_params, *, needs_reply=False, node_only=False):
    """Create the internal structure describing a command

    :param required_params: A list of parameters that must be included with the command.
    :param needs_reply: The message needs to be replied to (and must have a uuid).
    :param node_only: The message can only have originated from a node."""
    if 'user' in required_params or 'session' in required_params:
        raise RuntimeError('"user" and "session" are reserved parameter names')
    return required_params, needs_reply, node_only



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
from sys import stderr
from _thread import allocate_lock
from http.client import HTTPConnection
from base64 import b64decode, b64encode
from libnacl.public import SecretKey


class KeyPair:
    """Holds a public/secret key pair as base 64 - bytes not strings"""

    def __init__(self, name=None, prefix='~/.messidge', public=None, secret=None):
        # initialised the 'usual' way?
        self.public = public
        self.secret = secret
        if public is not None and secret is not None:
            return
        if name is None:
            return
        if prefix is None:
            raise RuntimeError("Prefix cannot be None")

        # we are fetching the keys
        expand = os.path.expanduser(prefix)
        try:
            with open(expand + '/' + name + ".pub", 'rb') as f:
                self.public = f.read()
        except FileNotFoundError:
            raise RuntimeError("No public key found, halting")
        try:
            with open(expand + '/' + name, 'rb') as f:
                self.secret = f.read()
        except FileNotFoundError:
            raise RuntimeError("No private key found, halting")

    def public_binary(self):
        return b64decode(self.public)

    def secret_binary(self):
        return b64decode(self.secret)

    @staticmethod
    def new(name=None, prefix='~/.messidge'):
        """Create a new random key pair"""
        keys = SecretKey()
        rtn = KeyPair()
        rtn.public = b64encode(keys.pk)
        rtn.secret = b64encode(keys.sk)

        # are we persisting them?
        if name is not None and prefix is not None:
            expand = os.path.expanduser(prefix)
            os.makedirs(expand, exist_ok=True)
            with open(expand + '/' + name + ".pub", 'w+b') as f:
                f.write(rtn.public)
            with open(expand + '/' + name, 'w+b') as f:
                f.write(rtn.secret)

        return rtn

    def __repr__(self):
        return "<messidge.Keys object at %x (pk=%s)>" % (id(self), self.public)


def default_location(prefix='~/.messidge'):
    try:
        with open(os.path.expanduser(prefix + '/default_location'), 'r') as f:
            return f.read().strip('\n')
    except FileNotFoundError:
        raise RuntimeError("""There is no default_location file so cannot choose default location.""")


def create_account(location, token, prefix='~/.messidge'):
    # The new key pair (will be written to disk)
    keys = KeyPair.new(name=location, prefix=prefix)

    # Let the server know
    conn = HTTPConnection(location, 2021)
    try:
        conn.request('POST', '/', token + " " + keys.public.decode())
    except ConnectionRefusedError:
        print("Server refused connection - please ensure the location address is correct", file=stderr)
        exit(1)
    reply = conn.getresponse().read().decode()

    # failed?
    if reply[:5] == 'Fail:':
        print(reply, file=stderr)
        exit(1)

    # write out the default location and server public key
    directory = os.path.expanduser(prefix) + '/'
    os.makedirs(directory, exist_ok=True)
    with open(directory + 'default_location', 'w') as f:
        f.write(location)
    with open(directory + location + '.spub', 'w') as f:
        f.write(reply)


class Waitable:
    def __init__(self, locked=True):
        self.wait_lock = allocate_lock()
        self.exception = None
        if locked:
            self.wait_lock.acquire()

    def __del__(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def wait_until_ready(self, timeout=120):
        """Blocks waiting for a (normally asynchronous) update indicating the object is ready.

        Note this also causes exceptions that would previously have been raised on the background thread
        to be raise on the calling (i.e. main) thread."""
        # this lock is used for waiting on while uploading layers, needs to be long
        self.wait_lock.acquire(timeout=timeout)
        self.wait_lock.release()
        if self.exception:
            raise self.exception
        return self

    def mark_as_ready(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def unblock_and_raise(self, exception):
        # releases the lock but causes the thread in 'wait_until_ready' to raise an exception
        self.exception = exception
        self.mark_as_ready()



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import time
import traceback
from _thread import get_ident

import zmq

from .client.message import Message


class Loop:

    def __init__(self, skt=None, message_type=Message, *, exit_on_exception=False):
        """Initialise (but not start) a message loop.

        :param skt: the (actually optional) ZMQ socket that connects to the location.
        :param message_type: optionally set to use a non-default message class.
        :param exit_on_exception: the message loop will exit if an exception makes it to the BaseException handler."""
        super().__init__()
        self.exclusive_handlers = {}
        self.reply_callbacks = {}
        self.command_handlers = {}
        self.idle = set()
        self.skt = skt  # the main trunk socket
        self.session_key = None
        self.message_type = message_type
        self.exit_on_exception = exit_on_exception
        self.value_error_handler = None
        self.running_thread = None
        self.caught_exception = None
        self.main_thread = get_ident()
        self.p = zmq.Poller()
        if skt is not None:
            self.p.register(skt, zmq.POLLIN)  # so register_reply will work even if we don't register anything else
            logging.debug("Message loop has registered trunk socket")

    def run(self):
        """Message loop. Runs single threaded (usually but not necessarily a background thread)."""
        self.running_thread = get_ident()
        last_idle = time.time()
        logging.debug("Message loop started")
        while self.running_thread is not None:
            try:
                # fetch the events
                events = self.p.poll(timeout=500)

                # idle? or every five seconds at worst
                tme = time.time()
                if len(events) == 0 or tme - last_idle > 5:
                    last_idle = tme
                    # should be be quitting?
                    if self.running_thread is None:  # yes, leave loop
                        break
                    # otherwise do our idle tasks
                    for idle_task in set(self.idle):
                        idle_task()

                # Deal with all the events
                for event in events:

                    # on an exclusive socket? (the actual process of collecting the message is owned by the callback)
                    socket = event[0]
                    if socket in self.exclusive_handlers:
                        self.exclusive_handlers[socket](socket)
                        continue

                    # maybe an fd socket hasn't disappeared yet?
                    if isinstance(socket, int):
                        logging.debug("Message arrived for an fd socket that was unregistered: " + str(socket))
                        continue

                    # an ordinary message
                    msg = self.message_type.receive(socket, self.session_key)

                    try:
                        # is this a hooked reply? - these may well deal with (or raise) exceptions...
                        try:
                            cb = self.reply_callbacks[msg.uuid]
                            if cb is not None:
                                cb(msg)
                            else:
                                # a blank reply callback means this is the background thread's result
                                # and we need to return control back to where it's blocking
                                logging.debug("Leaving nested message loop.")
                                return msg
                            continue
                        except KeyError:
                            pass

                        # non-main thread exceptions just bin out and don't take the app down
                        # so the message loop catches them and waits for the main thread to deal with it.
                        try:
                            logging.debug("Message loop caught an exception: " + msg.params["exception"])
                            raise ValueError(msg.params["exception"])
                        except (KeyError, TypeError):
                            pass

                        # hopefully, then, a vanilla command
                        try:
                            obj, handlers = self.command_handlers[socket]  # don't replace with single value
                        except KeyError:
                            logging.debug("Socket had left command handlers with a message still to process")
                            continue
                        try:
                            handler = handlers[msg.command]
                        except KeyError:
                            # logging.warning("No handler was found for: %s (uuid=%s)" % (msg.command, msg.uuid))
                            continue

                        # logging.debug("Handling command: " + str(msg.command))
                        Loop.check_basic_properties(msg, handler)
                        getattr(obj, '_' + msg.command.decode())(msg)

                    except ValueError as e:
                        if self.value_error_handler:
                            self.value_error_handler(e, msg)
                        else:
                            raise e

            except KeyboardInterrupt:
                self.stop()

            except BaseException as e:
                logging.warning(traceback.format_exc())
                self.caught_exception = e
                if self.exit_on_exception:
                    self.stop()

        logging.debug("Message loop has finished")

    def stop(self):
        """Causes the thread in 'run' to exit cleanly."""
        self.running_thread = None

    def set_crypto_params(self, session_key: bytes):
        """Sets the cryptographic parameters for this connection.

        :param session_key: The session key for this connection.
        """
        self.session_key = session_key

    def register_exclusive(self, obj, handler, *, comment: str=""):
        """Registers an object and handler called to receive all events received on a ZMQ socket or file descriptor.

        :param obj: the object that will receive events.
        :param handler: the handler that will be called - passed the zmq socket or file handler (int).
        :param comment: an optional comment for clarifying debug logs."""
        # function signature: def callback(self, socket)
        # object can be a zmq socket or a file descriptor
        if obj in self.exclusive_handlers:
            raise RuntimeError("Tried to register a socket exclusively twice: %s (%s)" % (str(obj), comment))
        if obj in self.command_handlers:
            raise RuntimeError("Socket is already registered with commands: %s (%s)" % (str(obj), comment))

        self.exclusive_handlers[obj] = handler
        self.p.register(obj, zmq.POLLIN)
        logging.debug("Message loop has registered exclusive: %s (%s)" % (str(obj), comment))

    def unregister_exclusive(self, obj):
        """Unregister an object as an exclusive handler.

        :param obj: The object to unregister."""
        try:
            del self.exclusive_handlers[obj]
            self.p.unregister(obj)
            logging.debug("Message loop has unregistered exclusive: " + str(obj))
        except KeyError:
            pass

    def register_commands(self, skt, obj, commands, *, comment: str=""):
        """Register command callbacks.

        :param skt: ZMQ socket that will receive the events.
        :param obj: The object that will handle the events.
        :param commands: A dictionary of {b'command': handler, ...}.
        :param comment: A string to help clarify debug logs."""
        if skt in self.exclusive_handlers:
            raise RuntimeError("Socket is already registered as exclusive: " + comment)
        if skt in self.command_handlers:
            raise RuntimeError("Tried to register a series of commands twice for the same socket: " + comment)
        for command in commands.keys():
            if not isinstance(command, bytes):
                raise RuntimeError("Pass commands as bytes, not strings")
        self.command_handlers[skt] = (obj, commands)
        if skt is not self.skt:  # OK to reclassify the existing socket but no OK to re-register it
            self.p.register(skt, zmq.POLLIN)

    def register_reply(self, command_uuid: bytes, callback):
        """Hooking the reply to a command. Note that this will not override an exclusive socket.

        :param command_uuid: the uuid of the command message from which we are expecting a reply.
        :param callback: the callback that gets sent the reply message when it arrives.

        Note that callback == None calls 'return' with the message"""
        self.reply_callbacks[command_uuid] = callback

    def unregister_reply(self, command_uuid: bytes):
        """Removing the hook for a command reply.

        :param command_uuid: the uuid of the command message from which we are no longer expecting a reply."""
        try:
            del self.reply_callbacks[command_uuid]
        except KeyError:
            logging.debug("Called unregister_reply for a uuid that isn't hooked")

    def register_on_idle(self, obj):
        """Register an object.method to be called whenever the message loop has idle time.

        :param obj: The object.method to be called - no additional parameters are passed."""
        # Idles get invoked every time the poll on the message loop times out (ie has nothing to do)
        if obj not in self.idle:
            self.idle.add(obj)

    def unregister_on_idle(self, obj):
        """Unregister an object.method from the message loop's idle time.

        :param obj: The object.method to be unregistered."""
        try:
            self.idle.remove(obj)
        except KeyError:
            pass

    def on_value_error(self, callback):
        """Register an alternative to raising exceptions for ValueError exceptions coming over the wire.

        :param callback: the callback to be fired, gets passed the exception and message that caused the exception."""
        self.value_error_handler = callback

    @staticmethod
    def check_basic_properties(msg, handler):
        """Helper utility to ensure messages fulfill certain basic criteria before they are passed to their handlers.

        :param msg: the message to be tested.
        :param handler: the handler (a return from client.connection.cmd) to be tested for."""
        necessary_params, needs_reply = handler
        for necessary in necessary_params:
            if necessary not in msg.params:
                raise ValueError("Necessary parameter was not passed: " + necessary)
        if needs_reply and not msg.replyable():
            raise ValueError("This command needs to be replyable but the message was not: " + str(msg))

    def __repr__(self):
        return "<messidge.loop.Loop object at %x (exclusive=%d commands=%d replies_callbacks=%d)>" % \
               (id(self), len(self.exclusive_handlers), len(self.command_handlers), len(self.reply_callbacks))



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import libnacl
import libnacl.utils
import os
import logging
from multiprocessing import Process, Pipe
from multiprocessing.connection import wait
from .message import BrokerMessage


class Agent(Process):
    """Subprocess async encryption"""

    def __init__(self):
        super().__init__(target=self.run)
        # create the sockets to send work up here
        self.encrypt_pipe = Pipe()
        self.decrypt_pipe = Pipe()
        self.stop_pipe = Pipe()

        # go
        self.running = False
        self.start()

    def encrypted_session_key(self, session_key, nonce, pk, secret_binary):
        return

    def stop(self):
        logging.debug("Stopping Agent")
        self.stop_pipe[0].send(b'')

    def run(self):
        # raise my priority
        os.setpriority(os.PRIO_PROCESS, 0, -15)

        # run a message loop
        self.running = True
        while self.running:
            try:
                ready_list = wait([self.encrypt_pipe[1], self.decrypt_pipe[1], self.stop_pipe[1]])
            except (KeyboardInterrupt, OSError):
                self._inner_stop()
                return

            for skt in ready_list:
                try:
                    if skt == self.decrypt_pipe[1]:
                        msg = BrokerMessage.receive_pipe(self.decrypt_pipe[1], True)
                        msg.decrypt()
                        msg.forward_pipe(self.decrypt_pipe[1])
                        continue
                except OSError:  # the pipe gets closed at an inappropriate moment
                    continue

                try:
                    if skt == self.encrypt_pipe[1]:
                        msg = BrokerMessage.receive_pipe(self.encrypt_pipe[1], False)
                        msg.encrypt()
                        msg.forward_pipe(self.encrypt_pipe[1])
                        continue
                except OSError:
                    continue

                if skt == self.stop_pipe[1]:
                    self._inner_stop()

    def _inner_stop(self):
        if not self.running:
            return
        self.decrypt_pipe[0].close()
        self.decrypt_pipe[1].close()
        self.encrypt_pipe[0].close()
        self.encrypt_pipe[1].close()
        self.stop_pipe[0].close()
        self.stop_pipe[1].close()
        self.running = False



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


import cbor
import libnacl
import libnacl.utils
import logging


class BrokerMessage:
    """The server side message, markedly different to it's client side equivalent"""
    # For instance, it actually has a destination

    def __init__(self):
        self.is_encrypted = None
        self.emit_pipe = None
        self.emit_socket = None
        self.bulk = None
        self.params = None
        self.rid = None
        self.nonce = None  # needed for decryption, a new one gets made when encrypting
        self.session_key = None  # ditto, except scoped to the entire session
        self.command = None
        self.uuid = None
        self.time = None

    @staticmethod
    def receive_socket(socket):
        # Receive a message from the socket.
        rid, binary = socket.recv_multipart()
        parts = cbor.loads(binary)

        rtn = BrokerMessage()
        rtn.rid = rid
        rtn.nonce = parts[0]
        rtn.command = parts[1]
        rtn.uuid = parts[2]
        rtn.params = parts[3]
        rtn.bulk = parts[4]
        rtn.is_encrypted = True  # unless it's an auth message, but auth deals with that
        rtn.session_key = None  # because we don't know it yet
        rtn.emit_pipe = None
        rtn.emit_socket = socket
        rtn.time = None
        # print("receive_socket " + str(rtn))
        return rtn

    @staticmethod
    def receive_pipe(pipe, encrypted=False, *, reply_through=None):
        # Receive a message from a multiprocessing pipe
        rid, session_key, parts = pipe.recv()

        rtn = BrokerMessage()
        rtn.rid = rid
        rtn.session_key = session_key
        rtn.nonce = parts[0]
        rtn.command = parts[1]
        rtn.uuid = parts[2]
        rtn.params = parts[3]
        rtn.bulk = parts[4]
        rtn.is_encrypted = encrypted
        rtn.emit_pipe = reply_through
        rtn.emit_socket = None
        rtn.time = None
        # print("receive_pipe " + str(rtn))
        return rtn

    def replyable(self):
        # Returns true if this message can be replied to (has a uuid)
        return self.uuid != b''

    def encrypt(self):
        if self.is_encrypted:
            return
        self.nonce = libnacl.utils.rand_nonce()
        params_binary = cbor.dumps(self.params)
        self.params = libnacl.crypto_secretbox(params_binary, self.nonce, self.session_key)
        if self.bulk != b'':
            self.bulk = libnacl.crypto_secretbox(self.bulk, self.nonce, self.session_key)
        self.is_encrypted = True

    def decrypt(self):
        if not self.is_encrypted:
            return
        try:
            params_string = libnacl.crypto_secretbox_open(self.params, self.nonce, self.session_key)
            self.params = cbor.loads(params_string)
            if self.bulk != b'':
                self.bulk = libnacl.crypto_secretbox_open(self.bulk, self.nonce, self.session_key)
            self.is_encrypted = False
        except ValueError:
            logging.warning("Asked to decrypt a message but could not, session: " + str(self.rid))
            self.params = None
            self.bulk = b''

    def reply(self, results=None, bulk=None):
        # Called on an existing message, presumably a command to provide the results
        # You can store a message and call reply more than once
        if results is None:
            self.results = {}
        if bulk is not None:
            self.bulk = bulk

        if self.emit_pipe is not None:
            # print("reply send_pipe " + str(self))
            BrokerMessage.send_pipe(self.emit_pipe, self.rid, self.session_key,
                                    b'', b'', self.uuid, results, bulk=self.bulk)
            return

        if self.emit_socket is not None:
            # print("reply send_socket " + str(self))
            BrokerMessage.send_socket(self.emit_socket, self.rid, b'', b'', self.uuid, results, bulk=self.bulk)
            return

        raise RuntimeError("Attempted to reply to a message without setting either emit pipe or socket")

    def forward_socket(self, skt):
        # print("forward_socket " + str(self))
        BrokerMessage.send_socket(skt, self.rid, self.nonce, self.command, self.uuid, self.params, bulk=self.bulk)

    def forward_pipe(self, pipe):
        # print("forward_pipe " + str(self))
        BrokerMessage.send_pipe(pipe, self.rid, self.session_key,
                                self.nonce, self.command, self.uuid, self.params, bulk=self.bulk)

    @staticmethod
    def send_socket(socket, rid, nonce, command, uuid, params, bulk=b''):
        # Send a message to a client - easier to call send_cmd on the broker
        binary = cbor.dumps((nonce, command, uuid, params if params is not None else {}, bulk))
        socket.send_multipart((rid, binary))

    @staticmethod
    def send_pipe(pipe, rid, session_key, nonce, command, uuid, params, bulk=b''):
        pipe.send((rid, session_key, (nonce, command, uuid, params if params is not None else {}, bulk)))

    def __repr__(self):
        return "<broker.message.BrokerMessage object at %x (command=%s uuid=%s encrypted=%s)>" % \
               (id(self), self.command, self.uuid, self.is_encrypted)
